/**
* Submitted for verification at blockscout.com on 2021-02-07 11:01:36.125540Z
*/
// File: dev/library/ERC721/introspection/IERC165.sol

// SPDX-License-Identifier: UNLICENSED
pragma solidity ^0.7.4;

/**
 * @dev Interface of the ERC165 standard, as defined in the
 * [EIP](https://eips.ethereum.org/EIPS/eip-165).
 *
 * Implementers can declare support of contract interfaces, which can then be
 * queried by others (`ERC165Checker`).
 *
 * For an implementation, see `ERC165`.
 */
interface IERC165 {
    /**
     * @dev Returns true if this contract implements the interface defined by
     * `interfaceId`. See the corresponding
     * [EIP section](https://eips.ethereum.org/EIPS/eip-165#how-interfaces-are-identified)
     * to learn more about how these ids are created.
     *
     * This function call must use less than 30 000 gas.
     */
    function supportsInterface(bytes4 interfaceId) external view returns (bool);
}

// File: dev/library/ERC721/IERC721.sol

pragma solidity ^0.7.4;


/**
 * @dev Required interface of an ERC721 compliant contract.
 */
abstract contract IERC721 is IERC165 {
    event Transfer(address indexed from, address indexed to, uint256 indexed tokenId);
    event Approval(address indexed owner, address indexed approved, uint256 indexed tokenId);
    event ApprovalForAll(address indexed owner, address indexed operator, bool approved);

    /**
     * @dev Returns the number of NFTs in `owner`'s account.
     */
    function balanceOf(address owner) virtual public view returns (uint256 balance);

    /**
     * @dev Returns the owner of the NFT specified by `tokenId`.
     */
    function ownerOf(uint256 tokenId) virtual public view returns (address owner);

    /**
     * @dev Transfers a specific NFT (`tokenId`) from one account (`from`) to
     * another (`to`).
     *
     * 
     *
     * Requirements:
     * - `from`, `to` cannot be zero.
     * - `tokenId` must be owned by `from`.
     * - If the caller is not `from`, it must be have been allowed to move this
     * NFT by either `approve` or `setApproveForAll`.
     */
    function safeTransferFrom(address from, address to, uint256 tokenId) virtual public payable;
    /**
     * @dev Transfers a specific NFT (`tokenId`) from one account (`from`) to
     * another (`to`).
     *
     * Requirements:
     * - If the caller is not `from`, it must be approved to move this NFT by
     * either `approve` or `setApproveForAll`.
     */
    function transferFrom(address from, address to, uint256 tokenId) virtual public payable;
    function approve(address to, uint256 tokenId) virtual public;
    function getApproved(uint256 tokenId) virtual public view returns (address operator);

    function setApprovalForAll(address operator, bool _approved) virtual public;
    function isApprovedForAll(address owner, address operator) virtual public view returns (bool);
    function safeTransferFrom(address from, address to, uint256 tokenId, bytes memory data) virtual public payable;
}

// File: dev/library/ERC721/IERC721Receiver.sol

pragma solidity ^0.7.4;

/**
 * @title ERC721 token receiver interface
 * @dev Interface for any contract that wants to support safeTransfers
 * from ERC721 asset contracts.
 */
abstract contract IERC721Receiver {
    /**
     * @notice Handle the receipt of an NFT
     * @dev The ERC721 smart contract calls this function on the recipient
     * after a `safeTransfer`. This function MUST return the function selector,
     * otherwise the caller will revert the transaction. The selector to be
     * returned can be obtained as `this.onERC721Received.selector`. This
     * function MAY throw to revert and reject the transfer.
     * Note: the ERC721 contract address is always the message sender.
     * @param operator The address which called `safeTransferFrom` function
     * @param from The address which previously owned the token
     * @param tokenId The NFT identifier which is being transferred
     * @param data Additional data with no specified format
     * @return bytes4 `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`
     */
    function onERC721Received(address operator, address from, uint256 tokenId, bytes memory data)
    virtual public returns (bytes4);
}

// File: dev/library/math/SafeMath.sol

pragma solidity ^0.7.4;

/**
 * @dev Wrappers over Solidity's arithmetic operations with added overflow
 * checks.
 *
 * Arithmetic operations in Solidity wrap on overflow. This can easily result
 * in bugs, because programmers usually assume that an overflow raises an
 * error, which is the standard behavior in high level programming languages.
 * `SafeMath` restores this intuition by reverting the transaction when an
 * operation overflows.
 *
 * Using this library instead of the unchecked operations eliminates an entire
 * class of bugs, so it's recommended to use it always.
 */
library SafeMath {
    /**
     * @dev Returns the addition of two unsigned integers, reverting on
     * overflow.
     *
     * Counterpart to Solidity's `+` operator.
     *
     * Requirements:
     * - Addition cannot overflow.
     */
    function add(uint256 a, uint256 b) internal pure returns (uint256) {
        uint256 c = a + b;
        require(c >= a, "SafeMath: addition overflow");

        return c;
    }

    /**
     * @dev Returns the subtraction of two unsigned integers, reverting on
     * overflow (when the result is negative).
     *
     * Counterpart to Solidity's `-` operator.
     *
     * Requirements:
     * - Subtraction cannot overflow.
     */
    function sub(uint256 a, uint256 b) internal pure returns (uint256) {
        require(b <= a, "SafeMath: subtraction overflow");
        uint256 c = a - b;

        return c;
    }

    /**
     * @dev Returns the multiplication of two unsigned integers, reverting on
     * overflow.
     *
     * Counterpart to Solidity's `*` operator.
     *
     * Requirements:
     * - Multiplication cannot overflow.
     */
    function mul(uint256 a, uint256 b) internal pure returns (uint256) {
        // Gas optimization: this is cheaper than requiring 'a' not being zero, but the
        // benefit is lost if 'b' is also tested.
        // See: https://github.com/OpenZeppelin/openzeppelin-contracts/pull/522
        if (a == 0) {
            return 0;
        }

        uint256 c = a * b;
        require(c / a == b, "SafeMath: multiplication overflow");

        return c;
    }

    /**
     * @dev Returns the integer division of two unsigned integers. Reverts on
     * division by zero. The result is rounded towards zero.
     *
     * Counterpart to Solidity's `/` operator. Note: this function uses a
     * `revert` opcode (which leaves remaining gas untouched) while Solidity
     * uses an invalid opcode to revert (consuming all remaining gas).
     *
     * Requirements:
     * - The divisor cannot be zero.
     */
    function div(uint256 a, uint256 b) internal pure returns (uint256) {
        // Solidity only automatically asserts when dividing by 0
        require(b > 0, "SafeMath: division by zero");
        uint256 c = a / b;
        // assert(a == b * c + a % b); // There is no case in which this doesn't hold

        return c;
    }

    /**
     * @dev Returns the remainder of dividing two unsigned integers. (unsigned integer modulo),
     * Reverts when dividing by zero.
     *
     * Counterpart to Solidity's `%` operator. This function uses a `revert`
     * opcode (which leaves remaining gas untouched) while Solidity uses an
     * invalid opcode to revert (consuming all remaining gas).
     *
     * Requirements:
     * - The divisor cannot be zero.
     */
    function mod(uint256 a, uint256 b) internal pure returns (uint256) {
        require(b != 0, "SafeMath: modulo by zero");
        return a % b;
    }
}

// File: dev/library/utils/Address.sol

pragma solidity ^0.7.4;

/**
 * @dev Collection of functions related to the address type,
 */
library Address {
    /**
     * @dev Returns true if `account` is a contract.
     *
     * This test is non-exhaustive, and there may be false-negatives: during the
     * execution of a contract's constructor, its address will be reported as
     * not containing a contract.
     *
     * > It is unsafe to assume that an address for which this function returns
     * false is an externally-owned account (EOA) and not a contract.
     */
    function isContract(address account) internal view returns (bool) {
        // This method relies in extcodesize, which returns 0 for contracts in
        // construction, since the code is only stored at the end of the
        // constructor execution.

        uint256 size;
        // solhint-disable-next-line no-inline-assembly
        assembly { size := extcodesize(account) }
        return size > 0;
    }
}

// File: dev/library/drafts/Counters.sol

pragma solidity ^0.7.4;


/**
 * @title Counters
 * @author Matt Condon (@shrugs)
 * @dev Provides counters that can only be incremented or decremented by one. This can be used e.g. to track the number
 * of elements in a mapping, issuing ERC721 ids, or counting request ids.
 *
 * Include with `using Counters for Counters.Counter;`
 * Since it is not possible to overflow a 256 bit integer with increments of one, `increment` can skip the SafeMath
 * overflow check, thereby saving gas. This does assume however correct usage, in that the underlying `_value` is never
 * directly accessed.
 */
library Counters {
    using SafeMath for uint256;

    struct Counter {
        // This variable should never be directly accessed by users of the library: interactions must be restricted to
        // the library's function. As of Solidity v0.5.2, this cannot be enforced, though there is a proposal to add
        // this feature: see https://github.com/ethereum/solidity/issues/4637
        uint256 _value; // default: 0
    }

    function current(Counter storage counter) internal view returns (uint256) {
        return counter._value;
    }

    function increment(Counter storage counter) internal {
        counter._value += 1;
    }

    function decrement(Counter storage counter) internal {
        counter._value = counter._value.sub(1);
    }
}

// File: dev/library/ERC721/introspection/ERC165.sol

pragma solidity ^0.7.4;


/**
 * @dev Implementation of the `IERC165` interface.
 *
 * Contracts may inherit from this and call `_registerInterface` to declare
 * their support of an interface.
 */
contract ERC165 is IERC165 {
    /*
     * bytes4(keccak256('supportsInterface(bytes4)')) == 0x01ffc9a7
     */
    bytes4 private constant _INTERFACE_ID_ERC165 = 0x01ffc9a7;

    /**
     * @dev Mapping of interface ids to whether or not it's supported.
     */
    mapping(bytes4 => bool) private _supportedInterfaces;

    constructor () {
        // Derived contracts need only register support for their own interfaces,
        // we register support for ERC165 itself here
        _registerInterface(_INTERFACE_ID_ERC165);
    }

    /**
     * @dev See `IERC165.supportsInterface`.
     *
     * Time complexity O(1), guaranteed to always use less than 30 000 gas.
     */
    function supportsInterface(bytes4 interfaceId) override external view returns (bool) {
        return _supportedInterfaces[interfaceId];
    }

    /**
     * @dev Registers the contract as an implementer of the interface defined by
     * `interfaceId`. Support of the actual ERC165 interface is automatic and
     * registering its interface id is not required.
     *
     * See `IERC165.supportsInterface`.
     *
     * Requirements:
     *
     * - `interfaceId` cannot be the ERC165 invalid interface (`0xffffffff`).
     */
    function _registerInterface(bytes4 interfaceId) internal {
        require(interfaceId != 0xffffffff, "ERC165: invalid interface id");
        _supportedInterfaces[interfaceId] = true;
    }
}

// File: dev/library/ERC721/ERC721.sol

pragma solidity ^0.7.4;







/**
 * @title ERC721 Non-Fungible Token Standard basic implementation
 * @dev see https://eips.ethereum.org/EIPS/eip-721
 */
contract ERC721 is ERC165, IERC721 {
    using SafeMath for uint256;
    using Address for address;
    using Counters for Counters.Counter;

    // Equals to `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`
    // which can be also obtained as `IERC721Receiver(0).onERC721Received.selector`
    bytes4 private constant _ERC721_RECEIVED = 0x150b7a02;

    // Mapping from token ID to owner
    mapping (uint256 => address) private _tokenOwner;

    // Mapping from token ID to approved address
    mapping (uint256 => address) private _tokenApprovals;

    // Mapping from owner to number of owned token
    mapping (address => Counters.Counter) private _ownedTokensCount;

    // Mapping from owner to operator approvals
    mapping (address => mapping (address => bool)) private _operatorApprovals;

    /*
     *     bytes4(keccak256('balanceOf(address)')) == 0x70a08231
     *     bytes4(keccak256('ownerOf(uint256)')) == 0x6352211e
     *     bytes4(keccak256('approve(address,uint256)')) == 0x095ea7b3
     *     bytes4(keccak256('getApproved(uint256)')) == 0x081812fc
     *     bytes4(keccak256('setApprovalForAll(address,bool)')) == 0xa22cb465
     *     bytes4(keccak256('isApprovedForAll(address,address)')) == 0xe985e9c
     *     bytes4(keccak256('transferFrom(address,address,uint256)')) == 0x23b872dd
     *     bytes4(keccak256('safeTransferFrom(address,address,uint256)')) == 0x42842e0e
     *     bytes4(keccak256('safeTransferFrom(address,address,uint256,bytes)')) == 0xb88d4fde
     *
     *     => 0x70a08231 ^ 0x6352211e ^ 0x095ea7b3 ^ 0x081812fc ^
     *        0xa22cb465 ^ 0xe985e9c ^ 0x23b872dd ^ 0x42842e0e ^ 0xb88d4fde == 0x80ac58cd
     */
    bytes4 private constant _INTERFACE_ID_ERC721 = 0x80ac58cd;

    constructor () {
        // register the supported interfaces to conform to ERC721 via ERC165
        _registerInterface(_INTERFACE_ID_ERC721);
    }

    /**
     * @dev Gets the balance of the specified address.
     * @param owner address to query the balance of
     * @return uint256 representing the amount owned by the passed address
     */
    function balanceOf(address owner) override public view returns (uint256) {
        require(owner != address(0), "ERC721: balance query for the zero address");

        return _ownedTokensCount[owner].current();
    }

    /**
     * @dev Gets the owner of the specified token ID.
     * @param tokenId uint256 ID of the token to query the owner of
     * @return address currently marked as the owner of the given token ID
     */
    function ownerOf(uint256 tokenId) override public view returns (address) {
        address owner = _tokenOwner[tokenId];
        require(owner != address(0), "ERC721: owner query for nonexistent token");

        return owner;
    }

    /**
     * @dev Approves another address to transfer the given token ID
     * The zero address indicates there is no approved address.
     * There can only be one approved address per token at a given time.
     * Can only be called by the token owner or an approved operator.
     * @param to address to be approved for the given token ID
     * @param tokenId uint256 ID of the token to be approved
     */
    function approve(address to, uint256 tokenId) override public {
        address owner = ownerOf(tokenId);
        require(to != owner, "ERC721: approval to current owner");

        require(msg.sender == owner || isApprovedForAll(owner, msg.sender),
            "ERC721: approve caller is not owner nor approved for all"
        );

        _tokenApprovals[tokenId] = to;
        emit Approval(owner, to, tokenId);
    }

    /**
     * @dev Gets the approved address for a token ID, or zero if no address set
     * Reverts if the token ID does not exist.
     * @param tokenId uint256 ID of the token to query the approval of
     * @return address currently approved for the given token ID
     */
    function getApproved(uint256 tokenId) override public view returns (address) {
        require(_exists(tokenId), "ERC721: approved query for nonexistent token");

        return _tokenApprovals[tokenId];
    }

    /**
     * @dev Sets or unsets the approval of a given operator
     * An operator is allowed to transfer all tokens of the sender on their behalf.
     * @param to operator address to set the approval
     * @param approved representing the status of the approval to be set
     */
    function setApprovalForAll(address to, bool approved) override public {
        require(to != msg.sender, "ERC721: approve to caller");

        _operatorApprovals[msg.sender][to] = approved;
        emit ApprovalForAll(msg.sender, to, approved);
    }

    /**
     * @dev Tells whether an operator is approved by a given owner.
     * @param owner owner address which you want to query the approval of
     * @param operator operator address which you want to query the approval of
     * @return bool whether the given operator is approved by the given owner
     */
    function isApprovedForAll(address owner, address operator) override public view returns (bool) {
        return _operatorApprovals[owner][operator];
    }

    /**
     * @dev Transfers the ownership of a given token ID to another address.
     * Usage of this method is discouraged, use `safeTransferFrom` whenever possible.
     * Requires the msg.sender to be the owner, approved, or operator.
     * @param from current owner of the token
     * @param to address to receive the ownership of the given token ID
     * @param tokenId uint256 ID of the token to be transferred
     */
    function transferFrom(address from, address to, uint256 tokenId) virtual override public payable{
        //solhint-disable-next-line max-line-length
        require(_isApprovedOrOwner(msg.sender, tokenId), "ERC721: transfer caller is not owner nor approved");

        _transferFrom(from, to, tokenId);
    }

    /**
     * @dev Safely transfers the ownership of a given token ID to another address
     * If the target address is a contract, it must implement `onERC721Received`,
     * which is called upon a safe transfer, and return the magic value
     * `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`; otherwise,
     * the transfer is reverted.
     * Requires the msg.sender to be the owner, approved, or operator
     * @param from current owner of the token
     * @param to address to receive the ownership of the given token ID
     * @param tokenId uint256 ID of the token to be transferred
     */
    function safeTransferFrom(address from, address to, uint256 tokenId) override public payable {
        safeTransferFrom(from, to, tokenId, "");
    }

    /**
     * @dev Safely transfers the ownership of a given token ID to another address
     * If the target address is a contract, it must implement `onERC721Received`,
     * which is called upon a safe transfer, and return the magic value
     * `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`; otherwise,
     * the transfer is reverted.
     * Requires the msg.sender to be the owner, approved, or operator
     * @param from current owner of the token
     * @param to address to receive the ownership of the given token ID
     * @param tokenId uint256 ID of the token to be transferred
     * @param _data bytes data to send along with a safe transfer check
     */
    function safeTransferFrom(address from, address to, uint256 tokenId, bytes memory _data) override public payable{
        transferFrom(from, to, tokenId);
        require(_checkOnERC721Received(from, to, tokenId, _data), "ERC721: transfer to non ERC721Receiver implementer");
    }

    /**
     * @dev Returns whether the specified token exists.
     * @param tokenId uint256 ID of the token to query the existence of
     * @return bool whether the token exists
     */
    function _exists(uint256 tokenId) internal view returns (bool) {
        address owner = _tokenOwner[tokenId];
        return owner != address(0);
    }

    /**
     * @dev Returns whether the given spender can transfer a given token ID.
     * @param spender address of the spender to query
     * @param tokenId uint256 ID of the token to be transferred
     * @return bool whether the msg.sender is approved for the given token ID,
     * is an operator of the owner, or is the owner of the token
     */
    function _isApprovedOrOwner(address spender, uint256 tokenId) internal view returns (bool) {
        require(_exists(tokenId), "ERC721: operator query for nonexistent token");
        address owner = ownerOf(tokenId);
        return (spender == owner || getApproved(tokenId) == spender || isApprovedForAll(owner, spender));
    }

    /**
     * @dev Internal function to mint a new token.
     * Reverts if the given token ID already exists.
     * @param to The address that will own the minted token
     * @param tokenId uint256 ID of the token to be minted
     */
    function _mint(address to, uint256 tokenId) virtual internal {
        require(to != address(0), "ERC721: mint to the zero address");
        require(!_exists(tokenId), "ERC721: token already minted");

        _tokenOwner[tokenId] = to;
        _ownedTokensCount[to].increment();

        emit Transfer(address(0), to, tokenId);
    }


    /**
     * @dev Internal function to transfer ownership of a given token ID to another address.
     * As opposed to transferFrom, this imposes no restrictions on msg.sender.
     * @param from current owner of the token
     * @param to address to receive the ownership of the given token ID
     * @param tokenId uint256 ID of the token to be transferred
     */
    function _transferFrom(address from, address to, uint256 tokenId) virtual internal {
        require(ownerOf(tokenId) == from, "ERC721: transfer of token that is not own");
        require(to != address(0), "ERC721: transfer to the zero address");

        _clearApproval(tokenId);

        _ownedTokensCount[from].decrement();
        _ownedTokensCount[to].increment();

        _tokenOwner[tokenId] = to;

        emit Transfer(from, to, tokenId);
    }

    /**
     * @dev Internal function to invoke `onERC721Received` on a target address.
     * The call is not executed if the target address is not a contract.
     *
     * This function is deprecated.
     * @param from address representing the previous owner of the given token ID
     * @param to target address that will receive the tokens
     * @param tokenId uint256 ID of the token to be transferred
     * @param _data bytes optional data to send along with the call
     * @return bool whether the call correctly returned the expected magic value
     */
    function _checkOnERC721Received(address from, address to, uint256 tokenId, bytes memory _data)
        internal returns (bool)
    {
        if (!to.isContract()) {
            return true;
        }

        bytes4 retval = IERC721Receiver(to).onERC721Received(msg.sender, from, tokenId, _data);
        return (retval == _ERC721_RECEIVED);
    }

    /**
     * @dev Private function to clear current approval of a given token ID.
     * @param tokenId uint256 ID of the token to be transferred
     */
    function _clearApproval(uint256 tokenId) private {
        if (_tokenApprovals[tokenId] != address(0)) {
            _tokenApprovals[tokenId] = address(0);
        }
    }
}

// File: dev/library/ERC721/IERC721Enumerable.sol

pragma solidity ^0.7.4;


/**
 * @title ERC-721 Non-Fungible Token Standard, optional enumeration extension
 * @dev See https://eips.ethereum.org/EIPS/eip-721
 */
abstract contract IERC721Enumerable is IERC721 {
    function totalSupply() virtual public view returns (uint256);
    function tokenOfOwnerByIndex(address owner, uint256 index) virtual public view returns (uint256 tokenId);

    function tokenByIndex(uint256 index) virtual public view returns (uint256);
}

// File: dev/library/ERC721/ERC721Enumerable.sol

pragma solidity ^0.7.4;




/**
 * @title ERC-721 Non-Fungible Token with optional enumeration extension logic
 * @dev See https://eips.ethereum.org/EIPS/eip-721
 */
contract ERC721Enumerable is ERC165, ERC721, IERC721Enumerable {
    // Mapping from owner to list of owned token IDs
    mapping(address => uint256[]) private _ownedTokens;

    // Mapping from token ID to index of the owner tokens list
    mapping(uint256 => uint256) private _ownedTokensIndex;

    using SafeMath for uint256;

    // Array with all token ids, used for enumeration
    uint256[] private _allTokens;

    // Mapping from token id to position in the allTokens array
    mapping(uint256 => uint256) private _allTokensIndex;

    /*
     *     bytes4(keccak256('totalSupply()')) == 0x18160ddd
     *     bytes4(keccak256('tokenOfOwnerByIndex(address,uint256)')) == 0x2f745c59
     *     bytes4(keccak256('tokenByIndex(uint256)')) == 0x4f6ccce7
     *
     *     => 0x18160ddd ^ 0x2f745c59 ^ 0x4f6ccce7 == 0x780e9d63
     */
    bytes4 private constant _INTERFACE_ID_ERC721_ENUMERABLE = 0x780e9d63;

    /**
     * @dev Constructor function.
     */
    constructor () {
        // register the supported interface to conform to ERC721Enumerable via ERC165
        _registerInterface(_INTERFACE_ID_ERC721_ENUMERABLE);
    }

    /**
     * @dev Gets the token ID at a given index of the tokens list of the requested owner.
     * @param owner address owning the tokens list to be accessed
     * @param index uint256 representing the index to be accessed of the requested tokens list
     * @return uint256 token ID at the given index of the tokens list owned by the requested address
     */
    function tokenOfOwnerByIndex(address owner, uint256 index) override public view returns (uint256) {
        require(index < balanceOf(owner), "ERC721Enumerable: owner index out of bounds");
        return _ownedTokens[owner][index];
    }

    /**
     * @dev Gets the total amount of tokens stored by the contract.
     * @return uint256 representing the total amount of tokens
     */
    function totalSupply() override public view returns (uint256) {
        return _allTokens.length;
    }

    /**
     * @dev Gets the token ID at a given index of all the tokens in this contract
     * Reverts if the index is greater or equal to the total number of tokens.
     * @param index uint256 representing the index to be accessed of the tokens list
     * @return uint256 token ID at the given index of the tokens list
     */
    function tokenByIndex(uint256 index) override public view returns (uint256) {
        require(index < totalSupply(), "ERC721Enumerable: global index out of bounds");
        return _allTokens[index];
    }

    /**
     * @dev Internal function to transfer ownership of a given token ID to another address.
     * As opposed to transferFrom, this imposes no restrictions on msg.sender.
     * @param from current owner of the token
     * @param to address to receive the ownership of the given token ID
     * @param tokenId uint256 ID of the token to be transferred
     */
    function _transferFrom(address from, address to, uint256 tokenId) virtual override internal {
        super._transferFrom(from, to, tokenId);

        _removeTokenFromOwnerEnumeration(from, tokenId);

        _addTokenToOwnerEnumeration(to, tokenId);
    }

    /**
     * @dev Internal function to mint a new token.
     * Reverts if the given token ID already exists.
     * @param to address the beneficiary that will own the minted token
     * @param tokenId uint256 ID of the token to be minted
     */
    function _mint(address to, uint256 tokenId) virtual override internal {
        super._mint(to, tokenId);

        _addTokenToOwnerEnumeration(to, tokenId);

        _addTokenToAllTokensEnumeration(tokenId);
    }

    /**
     * @dev Gets the list of token IDs of the requested owner.
     * @param owner address owning the tokens
     * @return uint256[] List of token IDs owned by the requested address
     */
    function _tokensOfOwner(address owner) internal view returns (uint256[] memory) {
        return _ownedTokens[owner];
    }

    function tokensOfOwner(address owner) public view returns (uint256[] memory) {
        return _tokensOfOwner(owner);
    }

    /**
     * @dev Private function to add a token to this extension's ownership-tracking data structures.
     * @param to address representing the new owner of the given token ID
     * @param tokenId uint256 ID of the token to be added to the tokens list of the given address
     */
    function _addTokenToOwnerEnumeration(address to, uint256 tokenId) private {
        _ownedTokensIndex[tokenId] = _ownedTokens[to].length;
        _ownedTokens[to].push(tokenId);
    }

    /**
     * @dev Private function to add a token to this extension's token tracking data structures.
     * @param tokenId uint256 ID of the token to be added to the tokens list
     */
    function _addTokenToAllTokensEnumeration(uint256 tokenId) private {
        _allTokensIndex[tokenId] = _allTokens.length;
        _allTokens.push(tokenId);
    }

    /**
     * @dev Private function to remove a token from this extension's ownership-tracking data structures. Note that
     * while the token is not assigned a new owner, the _ownedTokensIndex mapping is _not_ updated: this allows for
     * gas optimizations e.g. when performing a transfer operation (avoiding double writes).
     * This has O(1) time complexity, but alters the order of the _ownedTokens array.
     * @param from address representing the previous owner of the given token ID
     * @param tokenId uint256 ID of the token to be removed from the tokens list of the given address
     */
    function _removeTokenFromOwnerEnumeration(address from, uint256 tokenId) private {
        // To prevent a gap in from's tokens array, we store the last token in the index of the token to delete, and
        // then delete the last slot (swap and pop).

        uint256 lastTokenIndex = _ownedTokens[from].length.sub(1);
        uint256 tokenIndex = _ownedTokensIndex[tokenId];

        // When the token to delete is the last token, the swap operation is unnecessary
        if (tokenIndex != lastTokenIndex) {
            uint256 lastTokenId = _ownedTokens[from][lastTokenIndex];

            _ownedTokens[from][tokenIndex] = lastTokenId; // Move the last token to the slot of the to-delete token
            _ownedTokensIndex[lastTokenId] = tokenIndex; // Update the moved token's index
        }

        // This also deletes the contents at the last position of the array
        //_ownedTokens[from].length--;
        delete _ownedTokens[from][_ownedTokens[from].length.sub(1)];

        // Note that _ownedTokensIndex[tokenId] hasn't been cleared: it still points to the old slot (now occupied by
        // lastTokenId, or just over the end of the array if the token was the last one).
    }

}

// File: dev/library/ERC721/IERC721Metadata.sol

pragma solidity ^0.7.4;


/**
 * @title ERC-721 Non-Fungible Token Standard, optional metadata extension
 * @dev See https://eips.ethereum.org/EIPS/eip-721
 */
abstract contract IERC721Metadata is IERC721 {
    function name() virtual external view returns (string memory);
    function symbol() virtual external view returns (string memory);
    function tokenURI(uint256 tokenId) virtual external view returns (string memory);
}

// File: dev/library/utils/Strings.sol

pragma solidity ^0.7.4;

library Strings {

    function uint2str(uint i) internal pure returns (string memory) {
        if (i == 0) return "0";
        uint j = i;
        uint len;
        while (j != 0){
            len++;
            j /= 10;
        }
        bytes memory bstr = new bytes(len);
        uint k = len - 1;
        while (i != 0){
            bstr[k--] = byte(uint8(48 + i % 10));
            i /= 10;
        }
        return string(bstr);
    }
}

// File: dev/library/ERC721/ERC721Metadata.sol

pragma solidity ^0.7.4;





contract ERC721Metadata is ERC165, ERC721, IERC721Metadata {
    // Token name
    string private _name;

    // Token symbol
    string private _symbol;

    // Base Token URI
    string private _baseTokenURI;

    // Optional mapping for token URIs
    mapping(uint256 => string) private _tokenURIs;

    /*
     *     bytes4(keccak256('name()')) == 0x06fdde03
     *     bytes4(keccak256('symbol()')) == 0x95d89b41
     *     bytes4(keccak256('tokenURI(uint256)')) == 0xc87b56dd
     *
     *     => 0x06fdde03 ^ 0x95d89b41 ^ 0xc87b56dd == 0x5b5e139f
     */
    bytes4 private constant _INTERFACE_ID_ERC721_METADATA = 0x5b5e139f;

    /**
     * @dev Constructor function
     */
    constructor (string memory tokenName, string memory tokenSymbol) {
        _name = tokenName;
        _symbol = tokenSymbol;

        // register the supported interfaces to conform to ERC721 via ERC165
        _registerInterface(_INTERFACE_ID_ERC721_METADATA);
    }

    /**
     * @dev Gets the token name.
     * @return string representing the token name
     */
    function name() override external view returns (string memory) {
        return _name;
    }

    /**
     * @dev Gets the token symbol.
     * @return string representing the token symbol
     */
    function symbol() override external view returns (string memory) {
        return _symbol;
    }

    // - `external` to `public` (Allow 'super')
    // - Concat baseURI with tokenURI
    function tokenURI(uint256 tokenId) override public view returns (string memory) {
        return string(abi.encodePacked(_baseTokenURI, _tokenURI(tokenId)));
    }

    /**
     * @dev Internal returns an URI for a given token ID.
     * Throws if the token ID does not exist. May return an empty string.
     * @param tokenId uint256 ID of the token to query
     */
    function _tokenURI(uint256 tokenId) internal view returns (string memory) {
        require(_exists(tokenId), "ERC721Metadata: URI query for nonexistent token");
        return Strings.uint2str(tokenId);
    }   

    /**
     * @dev Internal function to set the base token URI.
     * @param uri string base URI to assign
     */
    function _setBaseTokenURI(string memory uri) internal {
        _baseTokenURI = uri;
    }

    function getBaseTokenURI() public view returns(string memory){
        return _baseTokenURI;
    }

}

// File: dev/library/ERC721/ERC721Full.sol

pragma solidity ^0.7.4;




/**
 * @title Full ERC721 Token
 * @dev This implementation includes all the required and some optional functionality of the ERC721 standard
 * Moreover, it includes approve all functionality using operator terminology.
 *
 * See https://eips.ethereum.org/EIPS/eip-721
 */
contract ERC721Full is ERC721, ERC721Enumerable, ERC721Metadata {
    constructor (string memory name, string memory symbol) ERC721Metadata(name, symbol) {
        // solhint-disable-previous-line no-empty-blocks
    }

    function _mint(address to, uint256 tokenId) override(ERC721, ERC721Enumerable) internal {
        super._mint(to, tokenId);
    }
    
    function _transferFrom(address from, address to, uint256 tokenId) override(ERC721, ERC721Enumerable) internal {
        super._transferFrom(from, to, tokenId);
    }
}

// File: dev/library/ownership/OwnableAndPausable.sol

pragma solidity ^0.7.4;

/**
 * @dev Contract module which provides a basic access control mechanism, where
 * there is an account (an owner) that can be granted exclusive access to
 * specific functions.
 *
 * This module is used through inheritance. It will make available the modifier
 * `onlyOwner`, which can be aplied to your functions to restrict their use to
 * the owner.
 */
contract OwnableAndPausable {
    address public _owner;
    bool public _paused;

    event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);
    event Paused(address account);
    event Unpaused(address account);

    constructor () {
        _owner = msg.sender;
        _paused = false;
        emit OwnershipTransferred(address(0), _owner);
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyGenuino() {
        require(isOwner(), "Ownable: caller is not the owner");
        _;
    }

    modifier whenNotPaused() {
        require(!_paused, "Pausable: paused");
        _;
    }

    modifier whenPaused() {
        require(_paused, "Pausable: not paused");
        _;
    }
    /**
     * @dev Returns true if the caller is the current owner.
     */
    function isOwner() public view returns (bool) {
        return msg.sender == _owner;
    }

    function _transferOwnership(address newOwner) internal {
        require(newOwner != address(0), "Ownable: new owner is the zero address");
        emit OwnershipTransferred(_owner, newOwner);
        _owner = newOwner;
    }


    function paused() public view returns (bool) {
        return _paused;
    }
    
    function pause() public onlyGenuino whenNotPaused {
        _paused = true;
        emit Paused(msg.sender);
    }

    function unpause() public onlyGenuino whenPaused {
        _paused = false;
        emit Unpaused(msg.sender);
    }
}

// File: dev/GenuNFT_STAGE.sol


pragma solidity ^0.7.4;
pragma experimental ABIEncoderV2;



/// @title GenuNFT
/// @author Genuino
contract GenuNFT_STAGE is ERC721Full, OwnableAndPausable {
  /**
   * Data Structures and values
   */

  struct NftData{
    uint256 creationDate;
    string nfcSerialNumber;
    string rfidSerialNumber;
    string patchId;
  }

  struct TrackingStep{
    uint256 timestamp;
    string eventCode;
    string eventDescription;
    string location;
    string locationName;
    string publicData1;
    string publicData2;
  }

  uint256 internal _transferFee;
  uint256 internal _creationFeeETH;

  mapping(uint256 => TrackingStep[]) tokenTracking;

  NftData[] tokenDataStorage;

  //mapping da tokenId a BrandId
  mapping(uint256 => uint256) tokenCreatedBy;

  //Mapping da brandId a tokenIds[]
  mapping(uint256 => uint256[]) tokensCreatedByArray;

  mapping (address => bool) _brandWhiteList;
  mapping (uint256 => address) _brandAddresses;
  mapping (address => uint256) _brandIds;
  //_brandId = 0 means invalid brand
  uint256 totalBrands = 0;

  // mapping tra tokenId e bool
  mapping (uint256 => bool) _tokenTransferDisabled;

  // mapping tra nfcSerialNumber e tokenId
  mapping (string => uint256) _nfcSerialsNumber;

  // mapping tra rfidSerialNumber e tokenId
  mapping (string => uint256) _rfidSerialsNumber;

  // mapping tra patchId e tokenId
  mapping (string => uint256) _patchIds;

  uint256 tokenIdCounter;

  /**********************/

  /*
  * Events
  */
  event TokenCreated(uint256 indexed tokenId, address creatorAddress, string nfcSerialNumber);
  event TokenTransfered(uint256 indexed tokenId, address from, address to);
  event TokenInfoUpdated(uint256 indexed tokenId, address updatingAddress);
  event BrandCreated(uint256 indexed brandId, address brandAddress);
  event BrandModified(uint256 indexed brandId, address newBrandAddress);
  event TokenTransferUpdated(uint256 indexed tokenId, address tokenOwner, bool disabled);
  event TokenOwnershipTransfered(uint256 indexed tokenId, address fromAddress, address toAddress);
  event TrackingStepAdded(uint256 indexed tokenId, TrackingStep newStep);
  event TransferFeesUpdated(uint256 newETHFees);
  event CreationFeesUpdated(uint256 newETHFees);
  event FeeHasBeenWithdrawn(uint256 withdrawnBalanceETH, address toAddress);

  /*....*/

  /*********************/
  constructor() ERC721Full("GenuNFT_STAGE", "GNFT_S") {
    addBrandAddress(msg.sender); //Genuino è un brand
  }

  /*
  * Internal functions for substitude modifiers
  */

  /**
  * @notice Check if caller is the contract owner 
  */
  function _isGenuino() internal view {
    require(msg.sender == _owner);
  }

  /**
   * @notice Check if caller is the contract owner or a brand
   */
  function _isGenuinoOrBrand() internal view {
    require(msg.sender == _owner || _brandWhiteList[msg.sender]);
  }

  /**
   * @notice Check if caller is the token owner
   */
  function _isTokenOwner(uint256 tokenId) internal view {
    require(msg.sender == ownerOf(tokenId));
  }

  /**
   * @notice Check if caller is the owner or a creator branf
   */
  function _isGenuinoOrCreatorBrand(uint256 tokenId) internal view {
    require(_exists(tokenId), "ERC721: operator query for nonexistent token");
    bool isCreator = _brandIds[msg.sender] != 0 && tokenCreatedBy[tokenId] == _brandIds[msg.sender];
    require(msg.sender == _owner || isCreator);
  }

  function _isNotPaused() internal view {
    require(_paused == false);
  }

  /*
   * Fee functions
   */

  /**
   * @dev update transfer fee
   * @param newFee the new fee to set
   */
  function updateTransferFee(uint256 newFee) public {
    _isGenuino();
    _transferFee = newFee;
    emit TransferFeesUpdated(_transferFee);
  }
  
  /**
   * @dev update creation fee
   * @param newFeeETH the new fee to set
   */
  function updateCreationFee(uint256 newFeeETH) public {
    _isGenuino();
    _creationFeeETH = newFeeETH;
    emit CreationFeesUpdated(_creationFeeETH);
  }
    
  /**
   * @dev withdraw fees
   */
  function withdrawFees() public {
    _isGenuino();
    uint256 balanceETH = address(this).balance;
    msg.sender.transfer(balanceETH);
    emit FeeHasBeenWithdrawn(balanceETH, msg.sender);
  }

  /*
   * NFT Token functions
   */

  /**
   * @dev Transfers ownership of the contract to a new account (`newOwner`).
   * Can only be called by the current owner.
   */
  function transferOwnership(address newOwner) public {
    _isGenuino();
    _transferOwnership(newOwner);
    _updateBrandAddress(1, newOwner);
  }

  /**
   * @dev Mint new token
   * @param tokenData the nft data
   */ 
  function createToken(NftData memory tokenData) public payable {
    _isNotPaused();
    _isGenuinoOrBrand();
    require(msg.value == _creationFeeETH, "ETH sent are not enough to create the token.");
    tokenDataStorage.push(tokenData);
    uint256 newTokenId = tokenIdCounter;
    tokenIdCounter++;
    _mint(msg.sender, newTokenId);
    uint256 brandId = _brandIds[msg.sender];
    tokenCreatedBy[newTokenId] = brandId;
    tokensCreatedByArray[brandId].push(newTokenId);
    _tokenTransferDisabled[newTokenId] = false;
    _nfcSerialsNumber[tokenData.nfcSerialNumber] = newTokenId;
    _rfidSerialsNumber[tokenData.rfidSerialNumber] = newTokenId;
    _patchIds[tokenData.patchId] = newTokenId;
    emit TokenCreated(newTokenId, msg.sender, tokenData.nfcSerialNumber);
  }

  /**
   * @dev Mint new tokens in batch
   * @param tokensData the nft datas
   */ 
  function createTokensInBatch(NftData[] memory tokensData) public payable {
    _isNotPaused();
    _isGenuinoOrBrand();
    for (uint i = 0; i < tokensData.length; i++) {
      _createToken(tokensData[i]);
    }
  }

  /** 
   * @notice Internal function to create token
   */
  function _createToken(NftData memory tokenData) internal {
    createToken(tokenData);
  }

  /**
   * @dev Mint new token with an initial step
   * @param tokenData the nft data
   * @param newStep the new tracking step 
   */ 
  function createTokenWithStep(NftData memory tokenData, TrackingStep memory newStep) public payable {
    _isNotPaused();
    _isGenuinoOrBrand();
    _createToken(tokenData);
    _addTrackingStep(tokenIdCounter - 1, newStep);
  }

  // https://github.com/OpenZeppelin/openzeppelin-contracts/issues/1015
  /**
   * @dev transfer token 
   * @param from address
   * @param to address
   * @param tokenId the token id to transfer 
   */
  function transferFrom(address from, address to, uint256 tokenId) override public payable {
    _isNotPaused();
    _isTokenOwner(tokenId);
    require(!_tokenTransferDisabled[tokenId], "The token ownership transfer has been disabled by the token owner.");
    require(_isApprovedOrOwner(msg.sender, tokenId), "ERC721: transfer caller is not owner nor approved");
    require(msg.value == _transferFee, "ETH sent are not enough to transfer the token.");
    addTransferStep(tokenId);
    //IERC721 token = IERC721()
    _transferFrom(msg.sender, to, tokenId);
    emit TokenOwnershipTransfered(tokenId, from, to);
  }

  
  /**
   * @dev funzione che viene chiamata per aggiungere/aggiornare informazioni del token
   * @param tokenId the id of the token
   * @param tokenData Nft Data struct
   */
  function updateItemInfo(uint256 tokenId, NftData memory tokenData) public {
    _isNotPaused();
    _isGenuinoOrCreatorBrand(tokenId);
    tokenDataStorage[tokenId] = tokenData;
    emit TokenInfoUpdated(tokenId, msg.sender);
  }

  /**
   * @dev utente disabilita la possibilità di trasferire un token 
   * @param tokenId the token id to disable 
   */
  function disableTokenTransfer(uint256 tokenId) public {
    _isNotPaused();
    _isTokenOwner(tokenId);
    _tokenTransferDisabled[tokenId] = true;
    emit TokenTransferUpdated(tokenId, msg.sender, true);
  }

  /**
   * @dev utente riabilita la possibilità di trasferire un token 
   * @param tokenId the token id to resume 
   */
  function resumeTokenTransfer(uint256 tokenId) public {
    _isNotPaused();
    _isTokenOwner(tokenId);
    _tokenTransferDisabled[tokenId] = false;
    emit TokenTransferUpdated(tokenId, msg.sender, false);
  }

  /**
   * @dev check if the tokenId is trasferible
   * @param tokenId the token id to check
   */
  function isTrasferable(uint256 tokenId) public view returns(bool) {
    return !_tokenTransferDisabled[tokenId];
  }

  /**
   * @dev set base token URI
   * @param uri the new uri to set
   */
  function setBaseTokenURI(string memory uri) public {
    _isGenuino();
    _setBaseTokenURI(uri);
  }

  /*
   * Brand functions
   */

  /**
   * @param newBrandAddress address to update brand
   * @return total brands created
   */ 
  function addBrandAddress(address newBrandAddress) public returns(uint256) {
    _isGenuino();
    require(!_brandWhiteList[newBrandAddress], "A brand with this address already exists.");
    totalBrands += 1;
    _brandAddresses[totalBrands] = newBrandAddress;
    _brandWhiteList[newBrandAddress] = true;
    _brandIds[newBrandAddress] = totalBrands;
    emit BrandCreated(totalBrands, newBrandAddress);
    return totalBrands;
  }

  /**
   * @param brandId the id of the brand
   */ 
  function removeBrandAddress(uint256 brandId) public {
    _isGenuino();
    require(brandId != 1, "Genuino brand cannot be removed.");
    updateBrandAddress(brandId, address(0));
  }

  /**
   * @dev Internal function for updating brand address
   * @param brandId the id of the brand
   * @param newBrandAddress address to update brand
   */ 
  function updateBrandAddress(uint256 brandId, address newBrandAddress) public {
    _isGenuino();
    _updateBrandAddress(brandId, newBrandAddress);
  }

  /**
   * @param brandId the id of the brand
   * @param newBrandAddress address to update brand
   */ 
  function _updateBrandAddress(uint256 brandId, address newBrandAddress) internal {
    require(!_brandWhiteList[newBrandAddress], "A brand with this address already exists.");
    address oldAddress = _brandAddresses[brandId];
    _brandWhiteList[oldAddress] = false;
    _brandAddresses[brandId] = newBrandAddress;
    _brandIds[oldAddress] = 0;
    _brandIds[newBrandAddress] = brandId;
    if(newBrandAddress != address(0)) { //removeBrandAddress Case
      _brandWhiteList[newBrandAddress] = true;
    }
    emit BrandModified(brandId, newBrandAddress);
  }

  /*
   * Tracking Step functions
   */

  /**
   * @dev Aggiunge uno step durante il trasferimento del token
   * @param tokenId the token id
   */
  function addTransferStep(uint256 tokenId) internal {
    _isTokenOwner(tokenId);
    string memory location = ''; 
    string memory locationName = '';
    if (tokenTracking[tokenId].length > 0) {
      TrackingStep memory lastStep = tokenTracking[tokenId][tokenTracking[tokenId].length - 1];
      location = lastStep.location;
      locationName = lastStep.locationName;
    }
    TrackingStep memory newStep = TrackingStep(
      getTimestamp(),
      'UTO',
      '{en-US:"Token Ownership Updated", it-IT:"Cessione Proprieta"}',
      location,
      locationName,
      '',
      ''
    );
    tokenTracking[tokenId].push(newStep);
    emit TrackingStepAdded(tokenId, newStep);
  }

  /**
   * @dev aggiorna la tappa del token
   * @param tokenId the token id
   * @param newStep the tracking step to added 
   */
  function addTrackingStep(uint256 tokenId, TrackingStep memory newStep) public {
    _isNotPaused();
    _isGenuinoOrCreatorBrand(tokenId);
    tokenTracking[tokenId].push(newStep);
    emit TrackingStepAdded(tokenId, newStep);
  }

  /** 
   * @notice Internal function to add a new tracking step
   */
  function _addTrackingStep(uint256 tokenId, TrackingStep memory newStep) internal {
      addTrackingStep(tokenId, newStep);
  }

  /**
   * @dev aggiorna le tappe del token
   * @param tokenId the token id
   * @param newSteps the tracking steps to added 
   */
  function addTrackingStepsInBatch(uint tokenId, TrackingStep[] memory newSteps) public {
    _isNotPaused();
    _isGenuinoOrCreatorBrand(tokenId);
    for(uint i = 0; i < newSteps.length; i++) {
      _addTrackingStep(tokenId, newSteps[i]);
    }
  }

  /**
   * @dev Aggiunge un tracking step con la nuova posizione
   * @param tokenId the token id
   * @param location the new location
   * @param locationName the new location name
   * @param publicData the public data 
   */
  function updateItemLocation(uint256 tokenId, string memory location, string memory locationName, string memory publicData) public {
    _isNotPaused();
    _isTokenOwner(tokenId);
    TrackingStep memory newStep = TrackingStep(
        getTimestamp(),
        'UIL',
        '{ en-US:"Location Updated", it-IT:"Posizione Aggiornata"}',
        location,
        locationName,
        '',
        publicData
    );
    tokenTracking[tokenId].push(newStep);
    emit TrackingStepAdded(tokenId, newStep);
  }

  /*
   * Getters
   */

  /**
   * @dev get creation fee
   * @return creation fees
   */
  function getCreationFee() public view returns(uint256) {
    return _creationFeeETH;
  }
    
  /**
   * @dev get transfer fee
   * @return transfer fees
   */
  function getTransferFee() public view returns(uint256) {
    return _transferFee;
  }
  
  /**
   * @dev retrieves all tokens created by brandId 
   * @param brandId brand id
   */ 
  function getTokensCreatedBy(uint256 brandId) public view returns(uint256[] memory){
    return tokensCreatedByArray[brandId];
  }

  /**
   * @return all tokens data
   */
  function getAllTokenData() public view returns(NftData[] memory){
    return tokenDataStorage;
  }

  /**
   * @dev get token by tokenId
   * @param tokenId tokenId to search
   * @return token NftData
   */
  function getTokenById(uint256 tokenId) public view returns(NftData memory){
    require(_exists(tokenId), "ERC721: operator query for nonexistent token");
    return tokenDataStorage[tokenId];
  }

  /**
   * @dev get tracking steps by tokenId
   * @param tokenId tokenId to search
   * @return token TrackingStep[]
   */
  function getTokenTrackingStepById(uint256 tokenId) public view returns(TrackingStep[] memory){
    require(_exists(tokenId), "ERC721: operator query for nonexistent token");
    return tokenTracking[tokenId];
  }

  /**
   * @return numbers of brands created
   */ 
  function getBrandsNumber() public view returns(uint256){
    return totalBrands;
  }

  /**
   * @param brandId brand id
   * @return brandId address
   */
  function getBrandAddress(uint256 brandId) public view returns(address){
    return _brandAddresses[brandId];
  }

  /**
   * @return all brands addresses
   */
  function getAllBrands() public view returns(address[] memory) {
    address[] memory allBrands = new address[](totalBrands+1);
    for(uint i = 0 ; i <= totalBrands ; i++){
      allBrands[i] = _brandAddresses[i];
    }
    return allBrands;
  }

  /**
   * @dev ritorna il token associato ad un nfc seriale
   * @param nfcSerialNumber the nfc serial number of the token
   * @return NftData
   * @return token id 
   */
  function getTokenByNfc(string memory nfcSerialNumber) public view returns(NftData memory, uint256) {
    uint256 tokenId = _nfcSerialsNumber[nfcSerialNumber];
    return (tokenDataStorage[tokenId], tokenId);
  }

  /**
   * @dev ritorna il token associato ad un rfid seriale
   * @param rfidSerialNumber the rfid serial number of the token
   * @return NftData
   * @return token id 
   */
  function getTokenByRfid(string memory rfidSerialNumber) public view returns(NftData memory, uint256) {
    uint256 tokenId = _rfidSerialsNumber[rfidSerialNumber];
    return (tokenDataStorage[tokenId], tokenId);
  }

  /**
   * @dev ritorna il token associato ad una patch id
   * @param patchId the patch id of the token
   * @return NftData
   * @return token id 
   */
  function getTokenByPatchId(string memory patchId) public view returns(NftData memory, uint256) {
    uint256 tokenId = _patchIds[patchId];
    return (tokenDataStorage[tokenId], tokenId);
  }

  /**
   * @notice retrieve timestamp of the current block 
   */
  function getTimestamp() public view returns (uint256) {
    return block.timestamp;
  }
}
